from typing import Annotated, AsyncGenerator

import pytest
from autogen_core._cancellation_token import CancellationToken
from autogen_core.code_executor import ImportFromModule
from autogen_core.tools import (
    BaseStreamTool,
    FunctionTool,
    StaticStreamWorkbench,
    StaticWorkbench,
    TextResultContent,
    ToolResult,
    Workbench,
)
from pydantic import BaseModel


class StreamArgs(BaseModel):
    count: int


class StreamResult(BaseModel):
    final_count: int


class StreamItem(BaseModel):
    current: int


class StreamTool(BaseStreamTool[StreamArgs, StreamItem, StreamResult]):
    def __init__(self) -> None:
        super().__init__(
            args_type=StreamArgs,
            return_type=StreamResult,
            name="test_stream_tool",
            description="A test stream tool that counts up to a number.",
        )

    async def run(self, args: StreamArgs, cancellation_token: CancellationToken) -> StreamResult:
        # For the regular run method, just return the final result
        return StreamResult(final_count=args.count)

    async def run_stream(
        self, args: StreamArgs, cancellation_token: CancellationToken
    ) -> AsyncGenerator[StreamItem | StreamResult, None]:
        for i in range(1, args.count + 1):
            if cancellation_token.is_cancelled():
                break
            yield StreamItem(current=i)
        yield StreamResult(final_count=args.count)


class StreamToolWithError(BaseStreamTool[StreamArgs, StreamItem, StreamResult]):
    def __init__(self) -> None:
        super().__init__(
            args_type=StreamArgs,
            return_type=StreamResult,
            name="test_stream_tool_error",
            description="A test stream tool that raises an error.",
        )

    async def run(self, args: StreamArgs, cancellation_token: CancellationToken) -> StreamResult:
        # For the regular run method, just raise the error
        raise ValueError("Stream tool error")

    async def run_stream(
        self, args: StreamArgs, cancellation_token: CancellationToken
    ) -> AsyncGenerator[StreamItem | StreamResult, None]:
        yield StreamItem(current=1)
        raise ValueError("Stream tool error")


@pytest.mark.asyncio
async def test_static_workbench() -> None:
    def test_tool_func_1(x: Annotated[int, "The number to double."]) -> int:
        return x * 2

    def test_tool_func_2(x: Annotated[int, "The number to add 2."]) -> int:
        raise ValueError("This is a test error")  # Simulate an error

    test_tool_1 = FunctionTool(
        test_tool_func_1,
        name="test_tool_1",
        description="A test tool that doubles a number.",
        global_imports=[ImportFromModule(module="typing_extensions", imports=["Annotated"])],
    )
    test_tool_2 = FunctionTool(
        test_tool_func_2,
        name="test_tool_2",
        description="A test tool that adds 2 to a number.",
        global_imports=[ImportFromModule(module="typing_extensions", imports=["Annotated"])],
    )

    # Create a StaticWorkbench instance with the test tools.
    async with StaticWorkbench(tools=[test_tool_1, test_tool_2]) as workbench:
        # List tools
        tools = await workbench.list_tools()
        assert len(tools) == 2
        assert "description" in tools[0]
        assert "parameters" in tools[0]
        assert tools[0]["name"] == "test_tool_1"
        assert tools[0]["description"] == "A test tool that doubles a number."
        assert tools[0]["parameters"] == {
            "type": "object",
            "properties": {"x": {"type": "integer", "title": "X", "description": "The number to double."}},
            "required": ["x"],
            "additionalProperties": False,
        }
        assert "description" in tools[1]
        assert "parameters" in tools[1]
        assert tools[1]["name"] == "test_tool_2"
        assert tools[1]["description"] == "A test tool that adds 2 to a number."
        assert tools[1]["parameters"] == {
            "type": "object",
            "properties": {"x": {"type": "integer", "title": "X", "description": "The number to add 2."}},
            "required": ["x"],
            "additionalProperties": False,
        }

        # Call tools
        result_1 = await workbench.call_tool("test_tool_1", {"x": 5})
        assert result_1.name == "test_tool_1"
        assert result_1.result[0].type == "TextResultContent"
        assert result_1.result[0].content == "10"
        assert result_1.to_text() == "10"
        assert result_1.is_error is False

        # Call tool with error
        result_2 = await workbench.call_tool("test_tool_2", {"x": 5})
        assert result_2.name == "test_tool_2"
        assert result_2.result[0].type == "TextResultContent"
        assert result_2.result[0].content == "This is a test error"
        assert result_2.to_text() == "This is a test error"
        assert result_2.is_error is True

        # Save state.
        state = await workbench.save_state()
        assert state["type"] == "StaticWorkbenchState"
        assert "tools" in state
        assert len(state["tools"]) == 2

        # Dump config.
        config = workbench.dump_component()

    # Load the workbench from the config.
    async with Workbench.load_component(config) as new_workbench:
        # Load state.
        await new_workbench.load_state(state)

        # Verify that the tools are still available after loading the state.
        tools = await new_workbench.list_tools()
        assert len(tools) == 2
        assert "description" in tools[0]
        assert "parameters" in tools[0]
        assert tools[0]["name"] == "test_tool_1"
        assert tools[0]["description"] == "A test tool that doubles a number."
        assert tools[0]["parameters"] == {
            "type": "object",
            "properties": {"x": {"type": "integer", "title": "X", "description": "The number to double."}},
            "required": ["x"],
            "additionalProperties": False,
        }
        assert "description" in tools[1]
        assert "parameters" in tools[1]
        assert tools[1]["name"] == "test_tool_2"
        assert tools[1]["description"] == "A test tool that adds 2 to a number."
        assert tools[1]["parameters"] == {
            "type": "object",
            "properties": {"x": {"type": "integer", "title": "X", "description": "The number to add 2."}},
            "required": ["x"],
            "additionalProperties": False,
        }

        # Call tools
        result_1 = await new_workbench.call_tool("test_tool_1", {"x": 5})
        assert result_1.name == "test_tool_1"
        assert result_1.result[0].type == "TextResultContent"
        assert result_1.result[0].content == "10"
        assert result_1.to_text() == "10"
        assert result_1.is_error is False

        # Call tool with error
        result_2 = await new_workbench.call_tool("test_tool_2", {"x": 5})
        assert result_2.name == "test_tool_2"
        assert result_2.result[0].type == "TextResultContent"
        assert result_2.result[0].content == "This is a test error"
        assert result_2.to_text() == "This is a test error"
        assert result_2.is_error is True


@pytest.mark.asyncio
async def test_static_stream_workbench_call_tool_stream() -> None:
    """Test call_tool_stream with streaming tools and regular tools."""

    def regular_tool_func(x: Annotated[int, "The number to double."]) -> int:
        return x * 2

    regular_tool = FunctionTool(
        regular_tool_func,
        name="regular_tool",
        description="A regular tool that doubles a number.",
        global_imports=[ImportFromModule(module="typing_extensions", imports=["Annotated"])],
    )

    stream_tool = StreamTool()
    stream_tool_with_error = StreamToolWithError()

    async with StaticStreamWorkbench(tools=[regular_tool, stream_tool, stream_tool_with_error]) as workbench:
        # Test streaming tool
        results: list[StreamItem | StreamResult | ToolResult] = []
        async for result in workbench.call_tool_stream("test_stream_tool", {"count": 3}):
            results.append(result)

        # Should get 3 intermediate results and 1 final result
        assert len(results) == 4

        # Check intermediate results (StreamItem objects)
        for i, result in enumerate(results[:3]):
            assert isinstance(result, StreamItem)
            assert result.current == i + 1

        # Check final result (ToolResult)
        final_result = results[-1]
        assert isinstance(final_result, ToolResult)
        assert final_result.name == "test_stream_tool"
        assert final_result.is_error is False
        assert final_result.result[0].type == "TextResultContent"
        assert "final_count" in final_result.result[0].content

        # Test regular (non-streaming) tool
        results_regular: list[ToolResult] = []
        async for result in workbench.call_tool_stream("regular_tool", {"x": 5}):
            results_regular.append(result)  # type: ignore

        # Should get only 1 result for non-streaming tool
        assert len(results_regular) == 1
        final_result = results_regular[0]
        assert final_result.name == "regular_tool"
        assert final_result.is_error is False
        assert final_result.result[0].content == "10"

        # Test streaming tool with error
        results_error: list[StreamItem | ToolResult] = []
        async for result in workbench.call_tool_stream("test_stream_tool_error", {"count": 3}):
            results_error.append(result)  # type: ignore

        # Should get 1 intermediate result and 1 error result
        assert len(results_error) == 2

        # Check intermediate result
        intermediate_result = results_error[0]
        assert isinstance(intermediate_result, StreamItem)
        assert intermediate_result.current == 1

        # Check error result
        error_result = results_error[1]
        assert isinstance(error_result, ToolResult)
        assert error_result.name == "test_stream_tool_error"
        assert error_result.is_error is True
        result_content = error_result.result[0]
        assert isinstance(result_content, TextResultContent)
        assert "Stream tool error" in result_content.content

        # Test tool not found
        results_not_found: list[ToolResult] = []
        async for result in workbench.call_tool_stream("nonexistent_tool", {"x": 5}):
            results_not_found.append(result)  # type: ignore

        assert len(results_not_found) == 1
        error_result = results_not_found[0]
        assert error_result.name == "nonexistent_tool"
        assert error_result.is_error is True
        result_content = error_result.result[0]
        assert isinstance(result_content, TextResultContent)
        assert "Tool nonexistent_tool not found" in result_content.content

        # Test with no arguments
        results_no_args: list[StreamItem | StreamResult | ToolResult] = []
        async for result in workbench.call_tool_stream("test_stream_tool", {"count": 1}):
            results_no_args.append(result)  # type: ignore

        assert len(results_no_args) == 2  # 1 intermediate + 1 final

        # Test with None arguments
        results_none: list[ToolResult] = []
        async for result in workbench.call_tool_stream("regular_tool", None):
            results_none.append(result)  # type: ignore

        # Should still work but may get error due to missing required argument
        assert len(results_none) == 1
        result = results_none[0]
        assert result.name == "regular_tool"
        # This should error because x is required
        assert result.is_error is True


@pytest.mark.asyncio
async def test_static_stream_workbench_call_tool_stream_cancellation() -> None:
    """Test call_tool_stream with cancellation token."""
    stream_tool = StreamTool()

    async with StaticStreamWorkbench(tools=[stream_tool]) as workbench:
        # Test with cancellation token
        cancellation_token = CancellationToken()

        results: list[StreamItem | StreamResult | ToolResult] = []
        async for result in workbench.call_tool_stream("test_stream_tool", {"count": 5}, cancellation_token):
            results.append(result)  # type: ignore
            if len(results) == 2:  # Cancel after 2 results
                cancellation_token.cancel()

        # Should get at least 2 results before cancellation
        assert len(results) >= 2


@pytest.mark.asyncio
async def test_static_stream_workbench_inheritance() -> None:
    """Test that StaticStreamWorkbench inherits from both StaticWorkbench and StreamWorkbench."""
    stream_tool = StreamTool()

    async with StaticStreamWorkbench(tools=[stream_tool]) as workbench:
        # Test that it has regular workbench functionality
        tools = await workbench.list_tools()
        assert len(tools) == 1
        assert tools[0]["name"] == "test_stream_tool"

        # Test regular call_tool method
        result = await workbench.call_tool("test_stream_tool", {"count": 2})
        assert result.name == "test_stream_tool"
        assert result.is_error is False

        # Test streaming functionality exists
        assert hasattr(workbench, "call_tool_stream")
        results: list[StreamItem | StreamResult | ToolResult] = []
        async for result in workbench.call_tool_stream("test_stream_tool", {"count": 2}):
            results.append(result)  # type: ignore
        assert len(results) == 3  # 2 intermediate + 1 final



import asyncio
import builtins
from typing import Any, AsyncGenerator, Dict, List, Literal, Mapping, Optional

from pydantic import BaseModel, Field
from typing_extensions import Self

from .._cancellation_token import CancellationToken
from .._component_config import Component, ComponentModel
from ._base import BaseTool, StreamTool, ToolOverride, ToolSchema
from ._workbench import StreamWorkbench, TextResultContent, ToolResult, Workbench


class StaticWorkbenchConfig(BaseModel):
    tools: List[ComponentModel] = []
    tool_overrides: Dict[str, ToolOverride] = Field(default_factory=dict)


class StateicWorkbenchState(BaseModel):
    type: Literal["StaticWorkbenchState"] = "StaticWorkbenchState"
    tools: Dict[str, Mapping[str, Any]] = {}


class StaticWorkbench(Workbench, Component[StaticWorkbenchConfig]):
    """
    A workbench that provides a static set of tools that do not change after
    each tool execution.

    Args:
        tools (List[BaseTool[Any, Any]]): A list of tools to be included in the workbench.
            The tools should be subclasses of :class:`~autogen_core.tools.BaseTool`.
        tool_overrides (Optional[Dict[str, ToolOverride]]): Optional mapping of original tool
            names to override configurations for name and/or description. This allows
            customizing how tools appear to consumers while maintaining the underlying
            tool functionality.
    """

    component_provider_override = "autogen_core.tools.StaticWorkbench"
    component_config_schema = StaticWorkbenchConfig

    def __init__(
        self, tools: List[BaseTool[Any, Any]], tool_overrides: Optional[Dict[str, ToolOverride]] = None
    ) -> None:
        self._tools = tools
        self._tool_overrides = tool_overrides or {}

        # Build reverse mapping from override names to original names for call_tool
        self._override_name_to_original: Dict[str, str] = {}
        existing_tool_names = {tool.name for tool in self._tools}

        for original_name, override in self._tool_overrides.items():
            if override.name and override.name != original_name:
                # Check for conflicts with existing tool names
                if override.name in existing_tool_names and override.name != original_name:
                    raise ValueError(
                        f"Tool override name '{override.name}' conflicts with existing tool name. "
                        f"Override names must not conflict with any tool names."
                    )
                # Check for conflicts with other override names
                if override.name in self._override_name_to_original:
                    existing_original = self._override_name_to_original[override.name]
                    raise ValueError(
                        f"Tool override name '{override.name}' is used by multiple tools: "
                        f"'{existing_original}' and '{original_name}'. Override names must be unique."
                    )
                self._override_name_to_original[override.name] = original_name

    async def list_tools(self) -> List[ToolSchema]:
        result_schemas: List[ToolSchema] = []
        for tool in self._tools:
            original_schema = tool.schema

            # Apply overrides if they exist for this tool
            if tool.name in self._tool_overrides:
                override = self._tool_overrides[tool.name]
                # Create a new ToolSchema with overrides applied
                schema: ToolSchema = {
                    "name": override.name if override.name is not None else original_schema["name"],
                    "description": override.description
                    if override.description is not None
                    else original_schema.get("description", ""),
                }
                # Copy optional fields
                if "parameters" in original_schema:
                    schema["parameters"] = original_schema["parameters"]
                if "strict" in original_schema:
                    schema["strict"] = original_schema["strict"]
            else:
                schema = original_schema

            result_schemas.append(schema)
        return result_schemas

    async def call_tool(
        self,
        name: str,
        arguments: Mapping[str, Any] | None = None,
        cancellation_token: CancellationToken | None = None,
        call_id: str | None = None,
    ) -> ToolResult:
        # Check if the name is an override name and map it back to the original
        original_name = self._override_name_to_original.get(name, name)

        tool = next((tool for tool in self._tools if tool.name == original_name), None)
        if tool is None:
            return ToolResult(
                name=name,  # Return the requested name (which might be overridden)
                result=[TextResultContent(content=f"Tool {name} not found.")],
                is_error=True,
            )
        if not cancellation_token:
            cancellation_token = CancellationToken()
        if not arguments:
            arguments = {}
        try:
            result_future = asyncio.ensure_future(tool.run_json(arguments, cancellation_token, call_id=call_id))
            cancellation_token.link_future(result_future)
            actual_tool_output = await result_future
            is_error = False
            result_str = tool.return_value_as_string(actual_tool_output)
        except Exception as e:
            result_str = self._format_errors(e)
            is_error = True
        return ToolResult(name=name, result=[TextResultContent(content=result_str)], is_error=is_error)

    async def start(self) -> None:
        return None

    async def stop(self) -> None:
        return None

    async def reset(self) -> None:
        return None

    async def save_state(self) -> Mapping[str, Any]:
        tool_states = StateicWorkbenchState()
        for tool in self._tools:
            tool_states.tools[tool.name] = await tool.save_state_json()
        return tool_states.model_dump()

    async def load_state(self, state: Mapping[str, Any]) -> None:
        parsed_state = StateicWorkbenchState.model_validate(state)
        for tool in self._tools:
            if tool.name in parsed_state.tools:
                await tool.load_state_json(parsed_state.tools[tool.name])

    def _to_config(self) -> StaticWorkbenchConfig:
        return StaticWorkbenchConfig(
            tools=[tool.dump_component() for tool in self._tools], tool_overrides=self._tool_overrides
        )

    @classmethod
    def _from_config(cls, config: StaticWorkbenchConfig) -> Self:
        return cls(tools=[BaseTool.load_component(tool) for tool in config.tools], tool_overrides=config.tool_overrides)

    def _format_errors(self, error: Exception) -> str:
        """Recursively format errors into a string."""

        error_message = ""
        if hasattr(builtins, "ExceptionGroup") and isinstance(error, builtins.ExceptionGroup):
            # ExceptionGroup is available in Python 3.11+.
            # TODO: how to make this compatible with Python 3.10?
            for sub_exception in error.exceptions:  # type: ignore
                error_message += self._format_errors(sub_exception)  # type: ignore
        else:
            error_message += f"{str(error)}\n"
        return error_message.strip()


class StaticStreamWorkbench(StaticWorkbench, StreamWorkbench):
    """
    A workbench that provides a static set of tools that do not change after
    each tool execution, and supports streaming results.
    """

    component_provider_override = "autogen_core.tools.StaticStreamWorkbench"

    async def call_tool_stream(
        self,
        name: str,
        arguments: Mapping[str, Any] | None = None,
        cancellation_token: CancellationToken | None = None,
        call_id: str | None = None,
    ) -> AsyncGenerator[Any | ToolResult, None]:
        tool = next((tool for tool in self._tools if tool.name == name), None)
        if tool is None:
            yield ToolResult(
                name=name,
                result=[TextResultContent(content=f"Tool {name} not found.")],
                is_error=True,
            )
            return
        if not cancellation_token:
            cancellation_token = CancellationToken()
        if not arguments:
            arguments = {}
        try:
            actual_tool_output: Any | None = None
            if isinstance(tool, StreamTool):
                previous_result: Any | None = None
                try:
                    async for result in tool.run_json_stream(arguments, cancellation_token, call_id=call_id):
                        if previous_result is not None:
                            yield previous_result
                        previous_result = result
                    actual_tool_output = previous_result
                except Exception as e:
                    # If there was a previous result before the exception, yield it first
                    if previous_result is not None:
                        yield previous_result
                    # Then yield the error result
                    result_str = self._format_errors(e)
                    yield ToolResult(name=tool.name, result=[TextResultContent(content=result_str)], is_error=True)
                    return
            else:
                # If the tool is not a stream tool, we run it normally and yield the result
                result_future = asyncio.ensure_future(tool.run_json(arguments, cancellation_token, call_id=call_id))
                cancellation_token.link_future(result_future)
                actual_tool_output = await result_future
            is_error = False
            result_str = tool.return_value_as_string(actual_tool_output)
        except Exception as e:
            result_str = self._format_errors(e)
            is_error = True
        yield ToolResult(name=tool.name, result=[TextResultContent(content=result_str)], is_error=is_error)



import datetime
from ipaddress import IPv4Address, IPv6Address
from typing import Annotated, Any, Dict, ForwardRef, List, Literal, Optional, Type, Union, cast

from pydantic import (
    UUID1,
    UUID3,
    UUID4,
    UUID5,
    AnyUrl,
    BaseModel,
    EmailStr,
    Field,
    Json,
    conbytes,
    confloat,
    conint,
    conlist,
    constr,
    create_model,
)
from pydantic.fields import FieldInfo


class SchemaConversionError(Exception):
    """Base class for schema conversion exceptions."""

    pass


class ReferenceNotFoundError(SchemaConversionError):
    """Raised when a $ref cannot be resolved."""

    pass


class FormatNotSupportedError(SchemaConversionError):
    """Raised when a format is not supported."""

    pass


class UnsupportedKeywordError(SchemaConversionError):
    """Raised when an unsupported JSON Schema keyword is encountered."""

    pass


TYPE_MAPPING: Dict[str, Type[Any]] = {
    "string": str,
    "integer": int,
    "boolean": bool,
    "number": float,
    "array": List,
    "object": dict,
    "null": type(None),
}

FORMAT_MAPPING: Dict[str, Any] = {
    "uuid": UUID4,
    "uuid1": UUID1,
    "uuid2": UUID4,
    "uuid3": UUID3,
    "uuid4": UUID4,
    "uuid5": UUID5,
    "email": EmailStr,
    "uri": AnyUrl,
    "hostname": constr(strict=True),
    "ipv4": IPv4Address,
    "ipv6": IPv6Address,
    "ipv4-network": IPv4Address,
    "ipv6-network": IPv6Address,
    "date-time": datetime.datetime,
    "date": datetime.date,
    "time": datetime.time,
    "duration": datetime.timedelta,
    "int32": conint(strict=True, ge=-(2**31), le=2**31 - 1),
    "int64": conint(strict=True, ge=-(2**63), le=2**63 - 1),
    "float": confloat(strict=True),
    "double": float,
    "decimal": float,
    "byte": conbytes(strict=True),
    "binary": conbytes(strict=True),
    "password": str,
    "path": str,
    "json": Json,
}


def _make_field(
    default: Any,
    *,
    title: Optional[str] = None,
    description: Optional[str] = None,
) -> Any:
    """Construct a Pydantic Field with proper typing."""
    field_kwargs: Dict[str, Any] = {}
    if title is not None:
        field_kwargs["title"] = title
    if description is not None:
        field_kwargs["description"] = description
    return Field(default, **field_kwargs)


class _JSONSchemaToPydantic:
    def __init__(self) -> None:
        self._model_cache: Dict[str, Optional[Union[Type[BaseModel], ForwardRef]]] = {}

    def _resolve_ref(self, ref: str, schema: Dict[str, Any]) -> Dict[str, Any]:
        ref_key = ref.split("/")[-1]
        definitions = cast(dict[str, dict[str, Any]], schema.get("$defs", {}))

        if ref_key not in definitions:
            raise ReferenceNotFoundError(
                f"Reference `{ref}` not found in `$defs`. Available keys: {list(definitions.keys())}"
            )

        return definitions[ref_key]

    def get_ref(self, ref_name: str) -> Any:
        if ref_name not in self._model_cache:
            raise ReferenceNotFoundError(
                f"Reference `{ref_name}` not found in cache. Available: {list(self._model_cache.keys())}"
            )

        if self._model_cache[ref_name] is None:
            return ForwardRef(ref_name)

        return self._model_cache[ref_name]

    def _process_definitions(self, root_schema: Dict[str, Any]) -> None:
        if "$defs" in root_schema:
            for model_name in root_schema["$defs"]:
                if model_name not in self._model_cache:
                    self._model_cache[model_name] = None

            for model_name, model_schema in root_schema["$defs"].items():
                if self._model_cache[model_name] is None:
                    self._model_cache[model_name] = self.json_schema_to_pydantic(model_schema, model_name, root_schema)

    def json_schema_to_pydantic(
        self, schema: Dict[str, Any], model_name: str = "GeneratedModel", root_schema: Optional[Dict[str, Any]] = None
    ) -> Type[BaseModel]:
        if root_schema is None:
            root_schema = schema
            self._process_definitions(root_schema)

        if "$ref" in schema:
            resolved = self._resolve_ref(schema["$ref"], root_schema)
            schema = {**resolved, **{k: v for k, v in schema.items() if k != "$ref"}}

        if "allOf" in schema:
            merged: Dict[str, Any] = {"type": "object", "properties": {}, "required": []}
            for s in schema["allOf"]:
                part = self._resolve_ref(s["$ref"], root_schema) if "$ref" in s else s
                merged["properties"].update(part.get("properties", {}))
                merged["required"].extend(part.get("required", []))
            for k, v in schema.items():
                if k not in {"allOf", "properties", "required"}:
                    merged[k] = v
            merged["required"] = list(set(merged["required"]))
            schema = merged

        return self._json_schema_to_model(schema, model_name, root_schema)

    def _resolve_union_types(self, schemas: List[Dict[str, Any]]) -> List[Any]:
        types: List[Any] = []
        for s in schemas:
            if "$ref" in s:
                types.append(self.get_ref(s["$ref"].split("/")[-1]))
            elif "enum" in s:
                types.append(Literal[tuple(s["enum"])] if len(s["enum"]) > 0 else Any)
            else:
                json_type = s.get("type")
                if json_type not in TYPE_MAPPING:
                    raise UnsupportedKeywordError(f"Unsupported or missing type `{json_type}` in union")

                # Handle array types with items specification
                if json_type == "array" and "items" in s:
                    item_schema = s["items"]
                    if "$ref" in item_schema:
                        item_type = self.get_ref(item_schema["$ref"].split("/")[-1])
                    else:
                        item_type_name = item_schema.get("type")
                        if item_type_name is None:
                            item_type = str
                        elif item_type_name not in TYPE_MAPPING:
                            raise UnsupportedKeywordError(f"Unsupported item type `{item_type_name}` in union array")
                        else:
                            item_type = TYPE_MAPPING[item_type_name]

                    constraints = {}
                    if "minItems" in s:
                        constraints["min_length"] = s["minItems"]
                    if "maxItems" in s:
                        constraints["max_length"] = s["maxItems"]

                    array_type = conlist(item_type, **constraints) if constraints else List[item_type]  # type: ignore[valid-type]
                    types.append(array_type)
                else:
                    types.append(TYPE_MAPPING[json_type])
        return types

    def _extract_field_type(self, key: str, value: Dict[str, Any], model_name: str, root_schema: Dict[str, Any]) -> Any:
        json_type = value.get("type")
        if json_type not in TYPE_MAPPING:
            raise UnsupportedKeywordError(
                f"Unsupported or missing type `{json_type}` for field `{key}` in `{model_name}`"
            )

        base_type = TYPE_MAPPING[json_type]
        constraints: Dict[str, Any] = {}

        if json_type == "string":
            if "minLength" in value:
                constraints["min_length"] = value["minLength"]
            if "maxLength" in value:
                constraints["max_length"] = value["maxLength"]
            if "pattern" in value:
                constraints["pattern"] = value["pattern"]
            if constraints:
                base_type = constr(**constraints)

        elif json_type == "integer":
            if "minimum" in value:
                constraints["ge"] = value["minimum"]
            if "maximum" in value:
                constraints["le"] = value["maximum"]
            if "exclusiveMinimum" in value:
                constraints["gt"] = value["exclusiveMinimum"]
            if "exclusiveMaximum" in value:
                constraints["lt"] = value["exclusiveMaximum"]
            if constraints:
                base_type = conint(**constraints)

        elif json_type == "number":
            if "minimum" in value:
                constraints["ge"] = value["minimum"]
            if "maximum" in value:
                constraints["le"] = value["maximum"]
            if "exclusiveMinimum" in value:
                constraints["gt"] = value["exclusiveMinimum"]
            if "exclusiveMaximum" in value:
                constraints["lt"] = value["exclusiveMaximum"]
            if constraints:
                base_type = confloat(**constraints)

        elif json_type == "array":
            if "minItems" in value:
                constraints["min_length"] = value["minItems"]
            if "maxItems" in value:
                constraints["max_length"] = value["maxItems"]
            item_schema = value.get("items", {"type": "string"})
            if "$ref" in item_schema:
                item_type = self.get_ref(item_schema["$ref"].split("/")[-1])
            else:
                item_type_name = item_schema.get("type")
                if item_type_name is None:
                    item_type = str
                elif item_type_name not in TYPE_MAPPING:
                    raise UnsupportedKeywordError(
                        f"Unsupported or missing item type `{item_type_name}` for array field `{key}` in `{model_name}`"
                    )
                else:
                    item_type = TYPE_MAPPING[item_type_name]

            base_type = conlist(item_type, **constraints) if constraints else List[item_type]  # type: ignore[valid-type]

        if "format" in value:
            format_type = FORMAT_MAPPING.get(value["format"])
            if format_type is None:
                raise FormatNotSupportedError(f"Unknown format `{value['format']}` for `{key}` in `{model_name}`")
            if not isinstance(format_type, type):
                return format_type
            if not issubclass(format_type, str):
                return format_type
            return format_type

        return base_type

    def _json_schema_to_model(
        self, schema: Dict[str, Any], model_name: str, root_schema: Dict[str, Any]
    ) -> Type[BaseModel]:
        if "allOf" in schema:
            merged: Dict[str, Any] = {"type": "object", "properties": {}, "required": []}
            for s in schema["allOf"]:
                part = self._resolve_ref(s["$ref"], root_schema) if "$ref" in s else s
                merged["properties"].update(part.get("properties", {}))
                merged["required"].extend(part.get("required", []))
            for k, v in schema.items():
                if k not in {"allOf", "properties", "required"}:
                    merged[k] = v
            merged["required"] = list(set(merged["required"]))
            schema = merged

        fields: Dict[str, tuple[Any, FieldInfo]] = {}
        required_fields = set(schema.get("required", []))

        for key, value in schema.get("properties", {}).items():
            if "$ref" in value:
                ref_name = value["$ref"].split("/")[-1]
                field_type = self.get_ref(ref_name)
            elif "anyOf" in value:
                sub_models = self._resolve_union_types(value["anyOf"])
                field_type = Union[tuple(sub_models)]
            elif "oneOf" in value:
                sub_models = self._resolve_union_types(value["oneOf"])
                field_type = Union[tuple(sub_models)]
                if "discriminator" in value:
                    discriminator = value["discriminator"]["propertyName"]
                    field_type = Annotated[field_type, Field(discriminator=discriminator)]
            elif "enum" in value:
                field_type = Literal[tuple(value["enum"])]
            elif "allOf" in value:
                merged = {"type": "object", "properties": {}, "required": []}
                for s in value["allOf"]:
                    part = self._resolve_ref(s["$ref"], root_schema) if "$ref" in s else s
                    merged["properties"].update(part.get("properties", {}))
                    merged["required"].extend(part.get("required", []))
                for k, v in value.items():
                    if k not in {"allOf", "properties", "required"}:
                        merged[k] = v
                merged["required"] = list(set(merged["required"]))
                field_type = self._json_schema_to_model(merged, f"{model_name}_{key}", root_schema)
            elif value.get("type") == "object" and "properties" in value:
                field_type = self._json_schema_to_model(value, f"{model_name}_{key}", root_schema)
            else:
                field_type = self._extract_field_type(key, value, model_name, root_schema)

            if field_type is None:
                raise UnsupportedKeywordError(f"Unsupported or missing type for field `{key}` in `{model_name}`")

            default_value = value.get("default")
            is_required = key in required_fields

            if not is_required and default_value is None:
                field_type = Optional[field_type]

            field_args = {
                "default": default_value if not is_required else ...,
            }
            if "title" in value:
                field_args["title"] = value["title"]
            if "description" in value:
                field_args["description"] = value["description"]

            fields[key] = (
                field_type,
                _make_field(
                    default_value if not is_required else ...,
                    title=value.get("title"),
                    description=value.get("description"),
                ),
            )

        model: Type[BaseModel] = create_model(model_name, **cast(dict[str, Any], fields))
        model.model_rebuild()
        return model


def schema_to_pydantic_model(schema: Dict[str, Any], model_name: str = "GeneratedModel") -> Type[BaseModel]:
    """
    Convert a JSON Schema dictionary to a fully-typed Pydantic model.

    This function handles schema translation and validation logic to produce
    a Pydantic model.

    **Supported JSON Schema Features**

    - **Primitive types**: `string`, `integer`, `number`, `boolean`, `object`, `array`, `null`
    - **String formats**:
        - `email`, `uri`, `uuid`, `uuid1`, `uuid3`, `uuid4`, `uuid5`
        - `hostname`, `ipv4`, `ipv6`, `ipv4-network`, `ipv6-network`
        - `date`, `time`, `date-time`, `duration`
        - `byte`, `binary`, `password`, `path`
    - **String constraints**:
        - `minLength`, `maxLength`, `pattern`
    - **Numeric constraints**:
        - `minimum`, `maximum`, `exclusiveMinimum`, `exclusiveMaximum`
    - **Array constraints**:
        - `minItems`, `maxItems`, `items`
    - **Object schema support**:
        - `properties`, `required`, `title`, `description`, `default`
    - **Enums**:
        - Converted to Python `Literal` type
    - **Union types**:
        - `anyOf`, `oneOf` supported with optional `discriminator`
    - **Inheritance and composition**:
        - `allOf` merges multiple schemas into one model
    - **$ref and $defs resolution**:
        - Supports references to sibling definitions and self-referencing schemas

    .. code-block:: python

        from autogen_core.utils import schema_to_pydantic_model

        # Example 1: Simple user model
        schema = {
            "title": "User",
            "type": "object",
            "properties": {
                "name": {"type": "string"},
                "email": {"type": "string", "format": "email"},
                "age": {"type": "integer", "minimum": 0},
            },
            "required": ["name", "email"],
        }

        UserModel = schema_to_pydantic_model(schema)
        user = UserModel(name="Alice", email="alice@example.com", age=30)

    .. code-block:: python

        from autogen_core.utils import schema_to_pydantic_model

        # Example 2: Nested model
        schema = {
            "title": "BlogPost",
            "type": "object",
            "properties": {
                "title": {"type": "string"},
                "tags": {"type": "array", "items": {"type": "string"}},
                "author": {
                    "type": "object",
                    "properties": {"name": {"type": "string"}, "email": {"type": "string", "format": "email"}},
                    "required": ["name"],
                },
            },
            "required": ["title", "author"],
        }

        BlogPost = schema_to_pydantic_model(schema)


    .. code-block:: python

        from autogen_core.utils import schema_to_pydantic_model

        # Example 3: allOf merging with $refs
        schema = {
            "title": "EmployeeWithDepartment",
            "allOf": [{"$ref": "#/$defs/Employee"}, {"$ref": "#/$defs/Department"}],
            "$defs": {
                "Employee": {
                    "type": "object",
                    "properties": {"id": {"type": "string"}, "name": {"type": "string"}},
                    "required": ["id", "name"],
                },
                "Department": {
                    "type": "object",
                    "properties": {"department": {"type": "string"}},
                    "required": ["department"],
                },
            },
        }

        Model = schema_to_pydantic_model(schema)

    .. code-block:: python

        from autogen_core.utils import schema_to_pydantic_model

        # Example 4: Self-referencing (recursive) model
        schema = {
            "title": "Category",
            "type": "object",
            "properties": {
                "name": {"type": "string"},
                "subcategories": {"type": "array", "items": {"$ref": "#/$defs/Category"}},
            },
            "required": ["name"],
            "$defs": {
                "Category": {
                    "type": "object",
                    "properties": {
                        "name": {"type": "string"},
                        "subcategories": {"type": "array", "items": {"$ref": "#/$defs/Category"}},
                    },
                    "required": ["name"],
                }
            },
        }

        Category = schema_to_pydantic_model(schema)

    .. code-block:: python

        # Example 5: Serializing and deserializing with Pydantic

        from uuid import uuid4
        from pydantic import BaseModel, EmailStr, Field
        from typing import Optional, List, Dict, Any
        from autogen_core.utils import schema_to_pydantic_model


        class Address(BaseModel):
            street: str
            city: str
            zipcode: str


        class User(BaseModel):
            id: str
            name: str
            email: EmailStr
            age: int = Field(..., ge=18)
            address: Address


        class Employee(BaseModel):
            id: str
            name: str
            manager: Optional["Employee"] = None


        class Department(BaseModel):
            name: str
            employees: List[Employee]


        class ComplexModel(BaseModel):
            user: User
            extra_info: Optional[Dict[str, Any]] = None
            sub_items: List[Employee]


        # Convert ComplexModel to JSON schema
        complex_schema = ComplexModel.model_json_schema()

        # Rebuild a new Pydantic model from JSON schema
        ReconstructedModel = schema_to_pydantic_model(complex_schema, "ComplexModel")

        # Instantiate reconstructed model
        reconstructed = ReconstructedModel(
            user={
                "id": str(uuid4()),
                "name": "Alice",
                "email": "alice@example.com",
                "age": 30,
                "address": {"street": "123 Main St", "city": "Wonderland", "zipcode": "12345"},
            },
            sub_items=[{"id": str(uuid4()), "name": "Bob", "manager": {"id": str(uuid4()), "name": "Eve"}}],
        )

        print(reconstructed.model_dump())


    Args:
        schema (Dict[str, Any]): A valid JSON Schema dictionary.
        model_name (str, optional): The name of the root model. Defaults to "GeneratedModel".

    Returns:
        Type[BaseModel]: A dynamically generated Pydantic model class.

    Raises:
        ReferenceNotFoundError: If a `$ref` key references a missing entry.
        FormatNotSupportedError: If a `format` keyword is unknown or unsupported.
        UnsupportedKeywordError: If the schema contains an unsupported `type`.

    See Also:
        - :class:`pydantic.BaseModel`
        - :func:`pydantic.create_model`
        - https://json-schema.org/
    """
    ...

    return _JSONSchemaToPydantic().json_schema_to_pydantic(schema, model_name)



import types
from typing import Any, Dict, List, Literal, Optional, Type, get_args, get_origin
from uuid import UUID, uuid4

import pytest
from autogen_core.utils._json_to_pydantic import (
    FORMAT_MAPPING,
    TYPE_MAPPING,
    FormatNotSupportedError,
    ReferenceNotFoundError,
    UnsupportedKeywordError,
    _JSONSchemaToPydantic,  # pyright: ignore[reportPrivateUsage]
)
from pydantic import BaseModel, EmailStr, Field, Json, ValidationError


# ✅ Define Pydantic models for testing
class Address(BaseModel):
    street: str
    city: str
    zipcode: str


class User(BaseModel):
    id: UUID
    name: str
    email: EmailStr
    age: int = Field(..., ge=18)  # Minimum age = 18
    address: Address


class Employee(BaseModel):
    id: UUID
    name: str
    manager: Optional["Employee"] = None  # Recursive self-reference


class Department(BaseModel):
    name: str
    employees: List[Employee]  # Array of objects


class ComplexModel(BaseModel):
    user: User
    extra_info: Optional[Dict[str, Any]] = None  # Optional dictionary
    sub_items: List[Employee]  # List of Employees
    json_string: Optional[Json[Any]] = None  # Optional JSON string


@pytest.fixture
def converter() -> _JSONSchemaToPydantic:
    """Fixture to create a fresh instance of JSONSchemaToPydantic for every test."""
    return _JSONSchemaToPydantic()


@pytest.fixture
def sample_json_schema() -> Dict[str, Any]:
    """Fixture that returns a JSON schema dynamically using model_json_schema()."""
    return User.model_json_schema()


@pytest.fixture
def sample_json_schema_recursive() -> Dict[str, Any]:
    """Fixture that returns a self-referencing JSON schema."""
    return Employee.model_json_schema()


@pytest.fixture
def sample_json_schema_nested() -> Dict[str, Any]:
    """Fixture that returns a nested schema with arrays of objects."""
    return Department.model_json_schema()


@pytest.fixture
def sample_json_schema_complex() -> Dict[str, Any]:
    """Fixture that returns a complex schema with multiple structures."""
    return ComplexModel.model_json_schema()


@pytest.mark.parametrize(
    "schema_fixture, model_name, expected_fields",
    [
        (sample_json_schema, "User", ["id", "name", "email", "age", "address"]),
        (sample_json_schema_recursive, "Employee", ["id", "name", "manager"]),
        (sample_json_schema_nested, "Department", ["name", "employees"]),
        (sample_json_schema_complex, "ComplexModel", ["user", "extra_info", "sub_items", "json_string"]),
    ],
)
def test_json_schema_to_pydantic(
    converter: _JSONSchemaToPydantic,
    schema_fixture: Any,
    model_name: str,
    expected_fields: List[str],
    request: Any,
) -> None:
    """Test conversion of JSON Schema to Pydantic model using the class instance."""
    schema = request.getfixturevalue(schema_fixture.__name__)
    Model = converter.json_schema_to_pydantic(schema, model_name)

    for field in expected_fields:
        assert field in Model.__annotations__, f"Expected '{field}' missing in {model_name}Model"


# ✅ **Valid Data Tests**
@pytest.mark.parametrize(
    "schema_fixture, model_name, valid_data",
    [
        (
            sample_json_schema,
            "User",
            {
                "id": str(uuid4()),
                "name": "Alice",
                "email": "alice@example.com",
                "age": 25,
                "address": {"street": "123 Main St", "city": "Metropolis", "zipcode": "12345"},
            },
        ),
        (
            sample_json_schema_recursive,
            "Employee",
            {
                "id": str(uuid4()),
                "name": "Alice",
                "manager": {
                    "id": str(uuid4()),
                    "name": "Bob",
                },
            },
        ),
        (
            sample_json_schema_nested,
            "Department",
            {
                "name": "Engineering",
                "employees": [
                    {
                        "id": str(uuid4()),
                        "name": "Alice",
                        "manager": {
                            "id": str(uuid4()),
                            "name": "Bob",
                        },
                    }
                ],
            },
        ),
        (
            sample_json_schema_complex,
            "ComplexModel",
            {
                "user": {
                    "id": str(uuid4()),
                    "name": "Charlie",
                    "email": "charlie@example.com",
                    "age": 30,
                    "address": {"street": "456 Side St", "city": "Gotham", "zipcode": "67890"},
                },
                "extra_info": {"hobby": "Chess", "level": "Advanced"},
                "sub_items": [
                    {"id": str(uuid4()), "name": "Eve"},
                    {"id": str(uuid4()), "name": "David", "manager": {"id": str(uuid4()), "name": "Frank"}},
                ],
                "json_string": '{"foo": "bar"}',
            },
        ),
    ],
)
def test_valid_data_model(
    converter: _JSONSchemaToPydantic,
    schema_fixture: Any,
    model_name: str,
    valid_data: Dict[str, Any],
    request: Any,
) -> None:
    """Test that valid data is accepted by the generated model."""
    schema = request.getfixturevalue(schema_fixture.__name__)
    Model = converter.json_schema_to_pydantic(schema, model_name)

    instance = Model(**valid_data)
    assert instance
    dumped = instance.model_dump(mode="json", exclude_none=True)
    assert dumped == valid_data, f"Model output mismatch.\nExpected: {valid_data}\nGot: {dumped}"


# ✅ **Invalid Data Tests**
@pytest.mark.parametrize(
    "schema_fixture, model_name, invalid_data",
    [
        (
            sample_json_schema,
            "User",
            {
                "id": "not-a-uuid",  # Invalid UUID
                "name": "Alice",
                "email": "not-an-email",  # Invalid email
                "age": 17,  # Below minimum
                "address": {"street": "123 Main St", "city": "Metropolis"},
            },
        ),
        (
            sample_json_schema_recursive,
            "Employee",
            {
                "id": str(uuid4()),
                "name": "Alice",
                "manager": {
                    "id": "not-a-uuid",  # Invalid UUID
                    "name": "Bob",
                },
            },
        ),
        (
            sample_json_schema_nested,
            "Department",
            {
                "name": "Engineering",
                "employees": [
                    {
                        "id": "not-a-uuid",  # Invalid UUID
                        "name": "Alice",
                        "manager": {
                            "id": str(uuid4()),
                            "name": "Bob",
                        },
                    }
                ],
            },
        ),
        (
            sample_json_schema_complex,
            "ComplexModel",
            {
                "user": {
                    "id": str(uuid4()),
                    "name": "Charlie",
                    "email": "charlie@example.com",
                    "age": "thirty",  # Invalid: Should be an int
                    "address": {"street": "456 Side St", "city": "Gotham", "zipcode": "67890"},
                },
                "extra_info": "should-be-dictionary",  # Invalid type
                "sub_items": [
                    {"id": "invalid-uuid", "name": "Eve"},  # Invalid UUID
                    {"id": str(uuid4()), "name": 123},  # Invalid name type
                ],
                "json_string": '{"foo": "bar"',  # Invalid JSON
            },
        ),
    ],
)
def test_invalid_data_model(
    converter: _JSONSchemaToPydantic,
    schema_fixture: Any,
    model_name: str,
    invalid_data: Dict[str, Any],
    request: Any,
) -> None:
    """Test that invalid data raises ValidationError."""
    schema = request.getfixturevalue(schema_fixture.__name__)
    Model = converter.json_schema_to_pydantic(schema, model_name)

    with pytest.raises(ValidationError):
        Model(**invalid_data)


class ListDictModel(BaseModel):
    """Example for `List[Dict[str, Any]]`"""

    data: List[Dict[str, Any]]


class DictListModel(BaseModel):
    """Example for `Dict[str, List[Any]]`"""

    mapping: Dict[str, List[Any]]


class NestedListModel(BaseModel):
    """Example for `List[List[str]]`"""

    matrix: List[List[str]]


@pytest.fixture
def sample_json_schema_list_dict() -> Dict[str, Any]:
    """Fixture for `List[Dict[str, Any]]`"""
    return ListDictModel.model_json_schema()


@pytest.fixture
def sample_json_schema_dict_list() -> Dict[str, Any]:
    """Fixture for `Dict[str, List[Any]]`"""
    return DictListModel.model_json_schema()


@pytest.fixture
def sample_json_schema_nested_list() -> Dict[str, Any]:
    """Fixture for `List[List[str]]`"""
    return NestedListModel.model_json_schema()


@pytest.mark.parametrize(
    "schema_fixture, model_name, expected_fields",
    [
        (sample_json_schema_list_dict, "ListDictModel", ["data"]),
        (sample_json_schema_dict_list, "DictListModel", ["mapping"]),
        (sample_json_schema_nested_list, "NestedListModel", ["matrix"]),
    ],
)
def test_json_schema_to_pydantic_nested(
    converter: _JSONSchemaToPydantic,
    schema_fixture: Any,
    model_name: str,
    expected_fields: list[str],
    request: Any,
) -> None:
    """Test conversion of JSON Schema to Pydantic model using the class instance."""
    schema = request.getfixturevalue(schema_fixture.__name__)
    Model = converter.json_schema_to_pydantic(schema, model_name)

    for field in expected_fields:
        assert field in Model.__annotations__, f"Expected '{field}' missing in {model_name}Model"


# ✅ **Valid Data Tests**
@pytest.mark.parametrize(
    "schema_fixture, model_name, valid_data",
    [
        (
            sample_json_schema_list_dict,
            "ListDictModel",
            {
                "data": [
                    {"key1": "value1", "key2": 10},
                    {"another_key": False, "nested": {"subkey": "data"}},
                ]
            },
        ),
        (
            sample_json_schema_dict_list,
            "DictListModel",
            {
                "mapping": {
                    "first": ["a", "b", "c"],
                    "second": [1, 2, 3, 4],
                    "third": [True, False, True],
                }
            },
        ),
        (
            sample_json_schema_nested_list,
            "NestedListModel",
            {"matrix": [["A", "B"], ["C", "D"], ["E", "F"]]},
        ),
    ],
)
def test_valid_data_model_nested(
    converter: _JSONSchemaToPydantic,
    schema_fixture: Any,
    model_name: str,
    valid_data: Dict[str, Any],
    request: Any,
) -> None:
    """Test that valid data is accepted by the generated model."""
    schema = request.getfixturevalue(schema_fixture.__name__)
    Model = converter.json_schema_to_pydantic(schema, model_name)

    instance = Model(**valid_data)
    assert instance
    for field, value in valid_data.items():
        assert (
            getattr(instance, field) == value
        ), f"Mismatch in field `{field}`: expected `{value}`, got `{getattr(instance, field)}`"


# ✅ **Invalid Data Tests**
@pytest.mark.parametrize(
    "schema_fixture, model_name, invalid_data",
    [
        (
            sample_json_schema_list_dict,
            "ListDictModel",
            {
                "data": "should-be-a-list",  # ❌ Should be a list of dicts
            },
        ),
        (
            sample_json_schema_dict_list,
            "DictListModel",
            {
                "mapping": [
                    "should-be-a-dictionary",  # ❌ Should be a dict of lists
                ]
            },
        ),
        (
            sample_json_schema_nested_list,
            "NestedListModel",
            {"matrix": [["A", "B"], "C", ["D", "E"]]},  # ❌ "C" is not a list
        ),
    ],
)
def test_invalid_data_model_nested(
    converter: _JSONSchemaToPydantic,
    schema_fixture: Any,
    model_name: str,
    invalid_data: Dict[str, Any],
    request: Any,
) -> None:
    """Test that invalid data raises ValidationError."""
    schema = request.getfixturevalue(schema_fixture.__name__)
    Model = converter.json_schema_to_pydantic(schema, model_name)

    with pytest.raises(ValidationError):
        Model(**invalid_data)


def test_reference_not_found(converter: _JSONSchemaToPydantic) -> None:
    schema = {"type": "object", "properties": {"manager": {"$ref": "#/$defs/MissingRef"}}}
    with pytest.raises(ReferenceNotFoundError):
        converter.json_schema_to_pydantic(schema, "MissingRefModel")


def test_format_not_supported(converter: _JSONSchemaToPydantic) -> None:
    schema = {"type": "object", "properties": {"custom_field": {"type": "string", "format": "unsupported-format"}}}
    with pytest.raises(FormatNotSupportedError):
        converter.json_schema_to_pydantic(schema, "UnsupportedFormatModel")


def test_unsupported_keyword(converter: _JSONSchemaToPydantic) -> None:
    schema = {"type": "object", "properties": {"broken_field": {"title": "Missing type"}}}
    with pytest.raises(UnsupportedKeywordError):
        converter.json_schema_to_pydantic(schema, "MissingTypeModel")


def test_enum_field_schema() -> None:
    schema = {
        "type": "object",
        "properties": {
            "status": {"type": "string", "enum": ["pending", "approved", "rejected"]},
            "priority": {"type": "integer", "enum": [1, 2, 3]},
        },
        "required": ["status"],
    }

    converter: _JSONSchemaToPydantic = _JSONSchemaToPydantic()
    Model = converter.json_schema_to_pydantic(schema, "Task")

    status_ann = Model.model_fields["status"].annotation
    assert get_origin(status_ann) is Literal
    assert set(get_args(status_ann)) == {"pending", "approved", "rejected"}

    priority_ann = Model.model_fields["priority"].annotation
    args = get_args(priority_ann)
    assert type(None) in args
    assert Literal[1, 2, 3] in args

    instance = Model(status="approved", priority=2)
    assert instance.status == "approved"  # type: ignore[attr-defined]
    assert instance.priority == 2  # type: ignore[attr-defined]


def test_metadata_title_description(converter: _JSONSchemaToPydantic) -> None:
    schema = {
        "title": "CustomerProfile",
        "description": "A profile containing personal and contact info",
        "type": "object",
        "properties": {
            "first_name": {"type": "string", "title": "First Name", "description": "Given name of the user"},
            "age": {"type": "integer", "title": "Age", "description": "Age in years"},
            "contact": {
                "type": "object",
                "title": "Contact Information",
                "description": "How to reach the user",
                "properties": {
                    "email": {
                        "type": "string",
                        "format": "email",
                        "title": "Email Address",
                        "description": "Primary email",
                    }
                },
            },
        },
        "required": ["first_name"],
    }

    Model: Type[BaseModel] = converter.json_schema_to_pydantic(schema, "CustomerProfile")
    generated_schema = Model.model_json_schema()

    assert generated_schema["title"] == "CustomerProfile"

    props = generated_schema["properties"]
    assert props["first_name"]["title"] == "First Name"
    assert props["first_name"]["description"] == "Given name of the user"
    assert props["age"]["title"] == "Age"
    assert props["age"]["description"] == "Age in years"

    contact = props["contact"]
    assert contact["title"] == "Contact Information"
    assert contact["description"] == "How to reach the user"

    # Follow the $ref
    ref_key = contact["anyOf"][0]["$ref"].split("/")[-1]
    contact_def = generated_schema["$defs"][ref_key]
    email = contact_def["properties"]["email"]
    assert email["title"] == "Email Address"
    assert email["description"] == "Primary email"


def test_oneof_with_discriminator(converter: _JSONSchemaToPydantic) -> None:
    schema = {
        "title": "PetWrapper",
        "type": "object",
        "properties": {
            "pet": {
                "oneOf": [{"$ref": "#/$defs/Cat"}, {"$ref": "#/$defs/Dog"}],
                "discriminator": {"propertyName": "pet_type"},
            }
        },
        "required": ["pet"],
        "$defs": {
            "Cat": {
                "type": "object",
                "properties": {"pet_type": {"type": "string", "enum": ["cat"]}, "hunting_skill": {"type": "string"}},
                "required": ["pet_type", "hunting_skill"],
                "title": "Cat",
            },
            "Dog": {
                "type": "object",
                "properties": {"pet_type": {"type": "string", "enum": ["dog"]}, "pack_size": {"type": "integer"}},
                "required": ["pet_type", "pack_size"],
                "title": "Dog",
            },
        },
    }

    Model = converter.json_schema_to_pydantic(schema, "PetWrapper")

    # Instantiate with a Cat
    cat = Model(pet={"pet_type": "cat", "hunting_skill": "expert"})
    assert cat.pet.pet_type == "cat"  # type: ignore[attr-defined]

    # Instantiate with a Dog
    dog = Model(pet={"pet_type": "dog", "pack_size": 4})
    assert dog.pet.pet_type == "dog"  # type: ignore[attr-defined]

    # Check round-trip schema includes discriminator
    model_schema = Model.model_json_schema()
    assert "discriminator" in model_schema["properties"]["pet"]
    assert model_schema["properties"]["pet"]["discriminator"]["propertyName"] == "pet_type"


def test_anyof_array_with_item_constraints(converter: _JSONSchemaToPydantic) -> None:
    """Test anyOf branch as array with items and min/max constraints."""
    schema = {
        "title": "ArrayOrString",
        "type": "object",
        "properties": {
            "value": {
                "anyOf": [
                    {
                        "type": "array",
                        "items": {"type": "string"},
                        "minItems": 1,
                        "maxItems": 3,
                    },
                    {"type": "string"},
                ]
            }
        },
        "required": ["value"],
    }

    Model = converter.json_schema_to_pydantic(schema, "ArrayOrString")

    m1 = Model(value="hello")
    assert m1.value == "hello"  # type: ignore[attr-defined]

    m2 = Model(value=["a", "b"])
    assert m2.value == ["a", "b"]  # type: ignore[attr-defined]

    with pytest.raises(ValidationError):
        Model(value=["one", "two", "three", "four"])

    with pytest.raises(ValidationError):
        Model(value=[])


def test_oneof_array_with_ref_items(converter: _JSONSchemaToPydantic) -> None:
    """Test oneOf branch as array whose items are a $ref object."""
    schema = {
        "title": "RefArrayOrInt",
        "type": "object",
        "properties": {
            "payload": {
                "oneOf": [
                    {
                        "type": "array",
                        "items": {"$ref": "#/$defs/Item"},
                        "minItems": 1,
                    },
                    {"type": "integer"},
                ]
            }
        },
        "required": ["payload"],
        "$defs": {
            "Item": {
                "type": "object",
                "properties": {"name": {"type": "string"}, "count": {"type": "integer"}},
                "required": ["name"],
                "title": "Item",
            }
        },
    }

    Model = converter.json_schema_to_pydantic(schema, "RefArrayOrInt")

    m1 = Model(payload=42)
    assert m1.payload == 42  # type: ignore[attr-defined]

    m2 = Model(payload=[{"name": "widget", "count": 5}, {"name": "gadget"}])
    assert [it.name for it in m2.payload] == ["widget", "gadget"]  # type: ignore[attr-defined]

    with pytest.raises(ValidationError):
        Model(payload=[])


def test_allof_merging_with_refs(converter: _JSONSchemaToPydantic) -> None:
    schema = {
        "title": "EmployeeWithDepartment",
        "allOf": [{"$ref": "#/$defs/Employee"}, {"$ref": "#/$defs/Department"}],
        "$defs": {
            "Employee": {
                "type": "object",
                "properties": {"id": {"type": "string"}, "name": {"type": "string"}},
                "required": ["id", "name"],
                "title": "Employee",
            },
            "Department": {
                "type": "object",
                "properties": {"department": {"type": "string"}},
                "required": ["department"],
                "title": "Department",
            },
        },
    }

    Model = converter.json_schema_to_pydantic(schema, "EmployeeWithDepartment")
    instance = Model(id="123", name="Alice", department="Engineering")
    assert instance.id == "123"  # type: ignore[attr-defined]
    assert instance.name == "Alice"  # type: ignore[attr-defined]
    assert instance.department == "Engineering"  # type: ignore[attr-defined]

    dumped = instance.model_dump()
    assert dumped == {"id": "123", "name": "Alice", "department": "Engineering"}


def test_nested_allof_merging(converter: _JSONSchemaToPydantic) -> None:
    schema = {
        "title": "ContainerModel",
        "type": "object",
        "properties": {
            "nested": {
                "type": "object",
                "properties": {
                    "data": {
                        "allOf": [
                            {"$ref": "#/$defs/Base"},
                            {"type": "object", "properties": {"extra": {"type": "string"}}, "required": ["extra"]},
                        ]
                    }
                },
                "required": ["data"],
            }
        },
        "required": ["nested"],
        "$defs": {
            "Base": {
                "type": "object",
                "properties": {"base_field": {"type": "string"}},
                "required": ["base_field"],
                "title": "Base",
            }
        },
    }

    Model = converter.json_schema_to_pydantic(schema, "ContainerModel")
    instance = Model(nested={"data": {"base_field": "abc", "extra": "xyz"}})

    assert instance.nested.data.base_field == "abc"  # type: ignore[attr-defined]
    assert instance.nested.data.extra == "xyz"  # type: ignore[attr-defined]


@pytest.mark.parametrize(
    "schema, field_name, valid_values, invalid_values",
    [
        # String constraints
        (
            {
                "type": "object",
                "properties": {
                    "username": {"type": "string", "minLength": 3, "maxLength": 10, "pattern": "^[a-zA-Z0-9_]+$"}
                },
                "required": ["username"],
            },
            "username",
            ["user_123", "abc", "Name2023"],
            ["", "ab", "toolongusername123", "invalid!char"],
        ),
        # Integer constraints
        (
            {
                "type": "object",
                "properties": {"age": {"type": "integer", "minimum": 18, "maximum": 99}},
                "required": ["age"],
            },
            "age",
            [18, 25, 99],
            [17, 100, -1],
        ),
        # Float constraints
        (
            {
                "type": "object",
                "properties": {"score": {"type": "number", "minimum": 0.0, "exclusiveMaximum": 1.0}},
                "required": ["score"],
            },
            "score",
            [0.0, 0.5, 0.999],
            [-0.1, 1.0, 2.5],
        ),
        # Array constraints
        (
            {
                "type": "object",
                "properties": {"tags": {"type": "array", "items": {"type": "string"}, "minItems": 1, "maxItems": 3}},
                "required": ["tags"],
            },
            "tags",
            [["a"], ["a", "b"], ["x", "y", "z"]],
            [[], ["one", "two", "three", "four"]],
        ),
    ],
)
def test_field_constraints(
    schema: Dict[str, Any],
    field_name: str,
    valid_values: List[Any],
    invalid_values: List[Any],
) -> None:
    converter = _JSONSchemaToPydantic()
    Model = converter.json_schema_to_pydantic(schema, "ConstraintModel")

    for value in valid_values:
        instance = Model(**{field_name: value})
        assert getattr(instance, field_name) == value

    for value in invalid_values:
        with pytest.raises(ValidationError):
            Model(**{field_name: value})


@pytest.mark.parametrize(
    "schema",
    [
        # Top-level field
        {"type": "object", "properties": {"weird": {"type": "abc"}}, "required": ["weird"]},
        # Inside array items
        {"type": "object", "properties": {"items": {"type": "array", "items": {"type": "abc"}}}, "required": ["items"]},
        # Inside anyOf
        {
            "type": "object",
            "properties": {"choice": {"anyOf": [{"type": "string"}, {"type": "abc"}]}},
            "required": ["choice"],
        },
    ],
)
def test_unknown_type_raises(schema: Dict[str, Any]) -> None:
    converter = _JSONSchemaToPydantic()
    with pytest.raises(UnsupportedKeywordError):
        converter.json_schema_to_pydantic(schema, "UnknownTypeModel")


@pytest.mark.parametrize("json_type, expected_type", list(TYPE_MAPPING.items()))
def test_basic_type_mapping(json_type: str, expected_type: type) -> None:
    schema = {
        "type": "object",
        "properties": {"field": {"type": json_type}},
        "required": ["field"],
    }
    converter = _JSONSchemaToPydantic()
    Model = converter.json_schema_to_pydantic(schema, f"{json_type.capitalize()}Model")

    assert "field" in Model.__annotations__
    field_type = Model.__annotations__["field"]

    # For array/object/null we check the outer type only
    if json_type == "null":
        assert field_type is type(None)
    elif json_type == "array":
        assert getattr(field_type, "__origin__", None) is list
    elif json_type == "object":
        assert field_type in (dict, Dict) or getattr(field_type, "__origin__", None) in (dict, Dict)

    else:
        assert field_type == expected_type


@pytest.mark.parametrize("format_name, expected_type", list(FORMAT_MAPPING.items()))
def test_format_mapping(format_name: str, expected_type: Any) -> None:
    schema = {
        "type": "object",
        "properties": {"field": {"type": "string", "format": format_name}},
        "required": ["field"],
    }
    converter = _JSONSchemaToPydantic()
    Model = converter.json_schema_to_pydantic(schema, f"{format_name.capitalize()}Model")

    assert "field" in Model.__annotations__
    field_type = Model.__annotations__["field"]
    if isinstance(expected_type, types.FunctionType):  # if it's a constrained constructor (e.g., conint)
        assert callable(field_type)
    else:
        assert field_type == expected_type


def test_unknown_format_raises() -> None:
    schema = {
        "type": "object",
        "properties": {"bad_field": {"type": "string", "format": "definitely-not-a-format"}},
    }
    converter = _JSONSchemaToPydantic()
    with pytest.raises(FormatNotSupportedError):
        converter.json_schema_to_pydantic(schema, "UnknownFormatModel")



from ._base import (
    BaseStreamTool,
    BaseTool,
    BaseToolWithState,
    ParametersSchema,
    StreamTool,
    Tool,
    ToolOverride,
    ToolSchema,
)
from ._function_tool import FunctionTool
from ._static_workbench import StaticStreamWorkbench, StaticWorkbench
from ._workbench import ImageResultContent, TextResultContent, ToolResult, Workbench

__all__ = [
    "Tool",
    "StreamTool",
    "ToolSchema",
    "ParametersSchema",
    "BaseTool",
    "BaseToolWithState",
    "BaseStreamTool",
    "FunctionTool",
    "Workbench",
    "ToolResult",
    "TextResultContent",
    "ImageResultContent",
    "StaticWorkbench",
    "StaticStreamWorkbench",
    "ToolOverride",
]



import pytest
from autogen_core.utils import extract_json_from_str


def test_extract_json_from_str() -> None:
    json_str = """
  {
    "name": "John",
    "age": 30,
    "city": "New York"
  }
  """
    json_resp = [{"name": "John", "age": 30, "city": "New York"}]
    resp = extract_json_from_str(json_str)
    assert resp == json_resp

    invalid_json_str = """
  {
    "name": "John",
    "age": 30,
    "city": "New York"
  """
    with pytest.raises(ValueError):
        extract_json_from_str(invalid_json_str)


def test_extract_json_from_str_codeblock() -> None:
    code_block_lang_str = """
  ```json
  {
    "name": "Alice",
    "age": 28,
    "city": "Seattle"
  }
  ```
  """
    code_block_no_lang_str = """
  ```
  {
    "name": "Alice",
    "age": 28,
    "city": "Seattle"
  }
  ```
  """
    code_block_resp = [{"name": "Alice", "age": 28, "city": "Seattle"}]
    multi_json_str = """
  ```json
  {
    "name": "John",
    "age": 30,
    "city": "New York"
  }
  ```
  ```json
  {
    "name": "Jane",
    "age": 25,
    "city": "Los Angeles"
  }
  ```
  """
    multi_json_resp = [
        {"name": "John", "age": 30, "city": "New York"},
        {"name": "Jane", "age": 25, "city": "Los Angeles"},
    ]

    lang_resp = extract_json_from_str(code_block_lang_str)
    assert lang_resp == code_block_resp
    no_lang_resp = extract_json_from_str(code_block_no_lang_str)
    assert no_lang_resp == code_block_resp
    multi_resp = extract_json_from_str(multi_json_str)
    assert multi_resp == multi_json_resp

    invalid_lang_code_block_str = """
  ```notjson
  {
    "name": "Jane",
    "age": 25,
    "city": "Los Angeles"
  }
  ```
  """
    with pytest.raises(ValueError):
        extract_json_from_str(invalid_lang_code_block_str)



import json
import re
from typing import Any, Dict, List


def extract_json_from_str(content: str) -> List[Dict[str, Any]]:
    """Extract JSON objects from a string. Supports backtick enclosed JSON objects"""
    pattern = re.compile(r"```(?:\s*([\w\+\-]+))?\n([\s\S]*?)```")
    matches = pattern.findall(content)
    ret: List[Dict[str, Any]] = []
    # If no matches found, assume the entire content is a JSON object
    if not matches:
        ret.append(json.loads(content))
    for match in matches:
        language = match[0].strip() if match[0] else None
        if language and language.lower() != "json":
            raise ValueError(f"Expected JSON object, but found language: {language}")
        content = match[1]
        ret.append(json.loads(content))
    return ret



from abc import ABC, abstractmethod
from types import TracebackType
from typing import Any, AsyncGenerator, List, Literal, Mapping, Optional, Type

from pydantic import BaseModel, Field
from typing_extensions import Annotated, Self

from .._cancellation_token import CancellationToken
from .._component_config import ComponentBase
from .._image import Image
from ._base import ToolSchema


class TextResultContent(BaseModel):
    """
    Text result content of a tool execution.
    """

    type: Literal["TextResultContent"] = "TextResultContent"

    content: str
    """The text content of the result."""


class ImageResultContent(BaseModel):
    """
    Image result content of a tool execution.
    """

    type: Literal["ImageResultContent"] = "ImageResultContent"

    content: Image
    """The image content of the result."""


ResultContent = Annotated[TextResultContent | ImageResultContent, Field(discriminator="type")]


class ToolResult(BaseModel):
    """
    A result of a tool execution by a workbench.
    """

    type: Literal["ToolResult"] = "ToolResult"

    name: str
    """The name of the tool that was executed."""

    result: List[ResultContent]
    """The result of the tool execution."""

    is_error: bool = False
    """Whether the tool execution resulted in an error."""

    def to_text(self, replace_image: str | None = None) -> str:
        """
        Convert the result to a text string.

        Args:
            replace_image (str | None): The string to replace the image content with.
                If None, the image content will be included in the text as base64 string.

        Returns:
            str: The text representation of the result.
        """
        parts: List[str] = []
        for content in self.result:
            if isinstance(content, TextResultContent):
                parts.append(content.content)
            elif isinstance(content, ImageResultContent):
                if replace_image is not None:
                    parts.append(replace_image)
                else:
                    parts.append(f"[Image: {content.content.to_base64()}]")
        return "\n".join(parts)


class Workbench(ABC, ComponentBase[BaseModel]):
    """
    A workbench is a component that provides a set of tools that may share
    resources and state.

    A workbench is responsible for managing the lifecycle of the tools and
    providing a single interface to call them. The tools provided by the workbench
    may be dynamic and their availabilities may change after each tool execution.

    A workbench can be started by calling the :meth:`~autogen_core.tools.Workbench.start` method
    and stopped by calling the :meth:`~autogen_core.tools.Workbench.stop` method.
    It can also be used as an asynchronous context manager, which will automatically
    start and stop the workbench when entering and exiting the context.
    """

    component_type = "workbench"

    @abstractmethod
    async def list_tools(self) -> List[ToolSchema]:
        """
        List the currently available tools in the workbench as :class:`ToolSchema`
        objects.

        The list of tools may be dynamic, and their content may change after
        tool execution.
        """
        ...

    @abstractmethod
    async def call_tool(
        self,
        name: str,
        arguments: Mapping[str, Any] | None = None,
        cancellation_token: CancellationToken | None = None,
        call_id: str | None = None,
    ) -> ToolResult:
        """
        Call a tool in the workbench.

        Args:
            name (str): The name of the tool to call.
            arguments (Mapping[str, Any] | None): The arguments to pass to the tool.
                If None, the tool will be called with no arguments.
            cancellation_token (CancellationToken | None): An optional cancellation token
                to cancel the tool execution.
            call_id (str | None): An optional identifier for the tool call, used for tracing.
        Returns:
            ToolResult: The result of the tool execution.
        """
        ...

    @abstractmethod
    async def start(self) -> None:
        """
        Start the workbench and initialize any resources.

        This method should be called before using the workbench.
        """
        ...

    @abstractmethod
    async def stop(self) -> None:
        """
        Stop the workbench and release any resources.

        This method should be called when the workbench is no longer needed.
        """
        ...

    @abstractmethod
    async def reset(self) -> None:
        """
        Reset the workbench to its initialized, started state.
        """
        ...

    @abstractmethod
    async def save_state(self) -> Mapping[str, Any]:
        """
        Save the state of the workbench.

        This method should be called to persist the state of the workbench.
        """
        ...

    @abstractmethod
    async def load_state(self, state: Mapping[str, Any]) -> None:
        """
        Load the state of the workbench.

        Args:
            state (Mapping[str, Any]): The state to load into the workbench.
        """
        ...

    async def __aenter__(self) -> Self:
        """
        Enter the workbench context manager.

        This method is called when the workbench is used in a `with` statement.
        It calls the :meth:`~autogen_core.tools.WorkBench.start` method to start the workbench.
        """
        await self.start()
        return self

    async def __aexit__(
        self, exc_type: Optional[Type[BaseException]], exc_val: Optional[BaseException], exc_tb: Optional[TracebackType]
    ) -> None:
        """
        Exit the workbench context manager.
        This method is called when the workbench is used in a `with` statement.
        It calls the :meth:`~autogen_core.tools.WorkBench.stop` method to stop the workbench.
        """
        await self.stop()


class StreamWorkbench(Workbench, ABC):
    """A workbench that supports streaming results from tool calls."""

    @abstractmethod
    def call_tool_stream(
        self,
        name: str,
        arguments: Mapping[str, Any] | None = None,
        cancellation_token: CancellationToken | None = None,
        call_id: str | None = None,
    ) -> AsyncGenerator[Any | ToolResult, None]:
        """
        Call a tool in the workbench and return a stream of results.

        Args:
            name (str): The name of the tool to call.
            arguments (Mapping[str, Any] | None): The arguments to pass to the tool
                If None, the tool will be called with no arguments.
            cancellation_token (CancellationToken | None): An optional cancellation token
                to cancel the tool execution.
            call_id (str | None): An optional identifier for the tool call, used for tracing.
        """
        ...



from typing import Annotated, Dict

import pytest
from autogen_core.code_executor import ImportFromModule
from autogen_core.tools import FunctionTool, StaticWorkbench, ToolOverride, Workbench


@pytest.mark.asyncio
async def test_static_workbench_with_tool_overrides() -> None:
    """Test StaticWorkbench with tool name and description overrides."""

    def test_tool_func_1(x: Annotated[int, "The number to double."]) -> int:
        return x * 2

    def test_tool_func_2(a: Annotated[int, "First number"], b: Annotated[int, "Second number"]) -> int:
        return a + b

    test_tool_1 = FunctionTool(
        test_tool_func_1,
        name="double",
        description="A test tool that doubles a number.",
        global_imports=[ImportFromModule(module="typing_extensions", imports=["Annotated"])],
    )
    test_tool_2 = FunctionTool(
        test_tool_func_2,
        name="add",
        description="A test tool that adds two numbers.",
        global_imports=[ImportFromModule(module="typing_extensions", imports=["Annotated"])],
    )

    # Define tool overrides
    overrides: Dict[str, ToolOverride] = {
        "double": ToolOverride(name="multiply_by_two", description="Multiplies a number by 2"),
        "add": ToolOverride(description="Performs addition of two integers"),  # Only override description
    }

    # Create a StaticWorkbench instance with tool overrides
    async with StaticWorkbench(tools=[test_tool_1, test_tool_2], tool_overrides=overrides) as workbench:
        # List tools and verify overrides are applied
        tools = await workbench.list_tools()
        assert len(tools) == 2

        # Check first tool has name and description overridden
        assert tools[0]["name"] == "multiply_by_two"
        assert tools[0].get("description") == "Multiplies a number by 2"
        assert tools[0].get("parameters") == {
            "type": "object",
            "properties": {"x": {"type": "integer", "title": "X", "description": "The number to double."}},
            "required": ["x"],
            "additionalProperties": False,
        }

        # Check second tool has only description overridden
        assert tools[1]["name"] == "add"  # Original name
        assert tools[1].get("description") == "Performs addition of two integers"  # Overridden description
        assert tools[1].get("parameters") == {
            "type": "object",
            "properties": {
                "a": {"type": "integer", "title": "A", "description": "First number"},
                "b": {"type": "integer", "title": "B", "description": "Second number"},
            },
            "required": ["a", "b"],
            "additionalProperties": False,
        }

        # Call tools using override names
        result_1 = await workbench.call_tool("multiply_by_two", {"x": 5})
        assert result_1.name == "multiply_by_two"  # Should return the override name
        assert result_1.result[0].type == "TextResultContent"
        assert result_1.result[0].content == "10"
        assert result_1.to_text() == "10"
        assert result_1.is_error is False

        # Call tool using original name (should still work for description-only override)
        result_2 = await workbench.call_tool("add", {"a": 3, "b": 7})
        assert result_2.name == "add"
        assert result_2.result[0].type == "TextResultContent"
        assert result_2.result[0].content == "10"
        assert result_2.to_text() == "10"
        assert result_2.is_error is False

        # Test calling non-existent tool
        result_3 = await workbench.call_tool("nonexistent", {"x": 5})
        assert result_3.name == "nonexistent"
        assert result_3.is_error is True
        assert result_3.result[0].type == "TextResultContent"
        assert "Tool nonexistent not found" in result_3.result[0].content


@pytest.mark.asyncio
async def test_static_workbench_without_overrides() -> None:
    """Test StaticWorkbench without overrides (original behavior)."""

    def test_tool_func(x: Annotated[int, "The number to double."]) -> int:
        return x * 2

    test_tool = FunctionTool(
        test_tool_func,
        name="double",
        description="A test tool that doubles a number.",
        global_imports=[ImportFromModule(module="typing_extensions", imports=["Annotated"])],
    )

    # Create workbench without overrides
    async with StaticWorkbench(tools=[test_tool]) as workbench:
        tools = await workbench.list_tools()
        assert len(tools) == 1
        assert tools[0].get("name") == "double"
        assert tools[0].get("description") == "A test tool that doubles a number."


@pytest.mark.asyncio
async def test_static_workbench_serialization_with_overrides() -> None:
    """Test that StaticWorkbench can be serialized and deserialized with overrides."""

    def test_tool_func(x: Annotated[int, "The number to double."]) -> int:
        return x * 2

    test_tool = FunctionTool(
        test_tool_func,
        name="double",
        description="A test tool that doubles a number.",
        global_imports=[ImportFromModule(module="typing_extensions", imports=["Annotated"])],
    )

    overrides: Dict[str, ToolOverride] = {
        "double": ToolOverride(name="multiply_by_two", description="Multiplies a number by 2")
    }

    # Create workbench with overrides
    workbench = StaticWorkbench(tools=[test_tool], tool_overrides=overrides)

    # Save configuration
    config = workbench.dump_component()
    assert "tool_overrides" in config.config

    # Load workbench from configuration
    async with Workbench.load_component(config) as new_workbench:
        tools = await new_workbench.list_tools()
        assert len(tools) == 1
        assert tools[0]["name"] == "multiply_by_two"
        assert tools[0].get("description") == "Multiplies a number by 2"

        # Test calling tool with override name
        result = await new_workbench.call_tool("multiply_by_two", {"x": 5})
        assert result.name == "multiply_by_two"
        assert result.result[0].content == "10"
        assert result.is_error is False


@pytest.mark.asyncio
async def test_static_workbench_partial_overrides() -> None:
    """Test StaticWorkbench with partial overrides (name only, description only)."""

    def tool1_func(x: Annotated[int, "Number"]) -> int:
        return x

    def tool2_func(x: Annotated[int, "Number"]) -> int:
        return x

    tool1 = FunctionTool(
        tool1_func,
        name="tool1",
        description="Original description 1",
        global_imports=[ImportFromModule(module="typing_extensions", imports=["Annotated"])],
    )
    tool2 = FunctionTool(
        tool2_func,
        name="tool2",
        description="Original description 2",
        global_imports=[ImportFromModule(module="typing_extensions", imports=["Annotated"])],
    )

    overrides: Dict[str, ToolOverride] = {
        "tool1": ToolOverride(name="renamed_tool1"),  # Only name override
        "tool2": ToolOverride(description="New description 2"),  # Only description override
    }

    async with StaticWorkbench(tools=[tool1, tool2], tool_overrides=overrides) as workbench:
        tools = await workbench.list_tools()

        # tool1: name overridden, description unchanged
        assert tools[0].get("name") == "renamed_tool1"
        assert tools[0].get("description") == "Original description 1"

        # tool2: name unchanged, description overridden
        assert tools[1].get("name") == "tool2"
        assert tools[1].get("description") == "New description 2"

        # Test calling with override name
        result1 = await workbench.call_tool("renamed_tool1", {"x": 42})
        assert result1.name == "renamed_tool1"
        assert result1.result[0].content == "42"

        # Test calling with original name
        result2 = await workbench.call_tool("tool2", {"x": 42})
        assert result2.name == "tool2"
        assert result2.result[0].content == "42"


def test_tool_override_model() -> None:
    """Test ToolOverride model functionality."""

    # Test with both fields
    override1 = ToolOverride(name="new_name", description="new_desc")
    assert override1.name == "new_name"
    assert override1.description == "new_desc"

    # Test with only name
    override2 = ToolOverride(name="new_name")
    assert override2.name == "new_name"
    assert override2.description is None

    # Test with only description
    override3 = ToolOverride(description="new_desc")
    assert override3.name is None
    assert override3.description == "new_desc"

    # Test empty
    override4 = ToolOverride()
    assert override4.name is None
    assert override4.description is None


def test_static_workbench_conflict_detection() -> None:
    """Test that StaticWorkbench detects conflicts in tool override names."""

    def test_tool_func_1(x: Annotated[int, "Number"]) -> int:
        return x

    def test_tool_func_2(x: Annotated[int, "Number"]) -> int:
        return x

    def test_tool_func_3(x: Annotated[int, "Number"]) -> int:
        return x

    tool1 = FunctionTool(
        test_tool_func_1,
        name="tool1",
        description="Tool 1",
        global_imports=[ImportFromModule(module="typing_extensions", imports=["Annotated"])],
    )
    tool2 = FunctionTool(
        test_tool_func_2,
        name="tool2",
        description="Tool 2",
        global_imports=[ImportFromModule(module="typing_extensions", imports=["Annotated"])],
    )
    tool3 = FunctionTool(
        test_tool_func_3,
        name="tool3",
        description="Tool 3",
        global_imports=[ImportFromModule(module="typing_extensions", imports=["Annotated"])],
    )

    # Test 1: Valid overrides - should work
    overrides_valid: Dict[str, ToolOverride] = {
        "tool1": ToolOverride(name="renamed_tool1"),
        "tool2": ToolOverride(name="renamed_tool2"),
    }
    workbench_valid = StaticWorkbench(tools=[tool1, tool2, tool3], tool_overrides=overrides_valid)
    assert "renamed_tool1" in workbench_valid._override_name_to_original  # type: ignore[reportPrivateUsage]
    assert "renamed_tool2" in workbench_valid._override_name_to_original  # type: ignore[reportPrivateUsage]

    # Test 2: Conflict with existing tool name - should fail
    overrides_conflict: Dict[str, ToolOverride] = {
        "tool1": ToolOverride(name="tool2")  # tool2 already exists
    }
    with pytest.raises(ValueError):
        StaticWorkbench(tools=[tool1, tool2, tool3], tool_overrides=overrides_conflict)

    # Test 3: Duplicate override names - should fail
    overrides_duplicate: Dict[str, ToolOverride] = {
        "tool1": ToolOverride(name="same_name"),
        "tool2": ToolOverride(name="same_name"),  # Duplicate
    }
    with pytest.raises(ValueError):
        StaticWorkbench(tools=[tool1, tool2, tool3], tool_overrides=overrides_duplicate)

    # Test 4: Self-renaming - should work but not add to reverse mapping
    overrides_self: Dict[str, ToolOverride] = {
        "tool1": ToolOverride(name="tool1")  # Renaming to itself
    }
    workbench_self = StaticWorkbench(tools=[tool1, tool2, tool3], tool_overrides=overrides_self)
    assert "tool1" not in workbench_self._override_name_to_original  # type: ignore[reportPrivateUsage]



from collections.abc import Generator
from contextlib import contextmanager
from enum import Enum
from typing import Any, Optional

from opentelemetry import trace
from opentelemetry.trace import Span, SpanKind

from .._agent_instantiation import AgentInstantiationContext

# OpenTelemetry semantic convention constants for GenAI operations
# Copied from opentelemetry-semantic-conventions to avoid dependency

# GenAI Agent attributes
GEN_AI_AGENT_DESCRIPTION = "gen_ai.agent.description"
GEN_AI_AGENT_ID = "gen_ai.agent.id"
GEN_AI_AGENT_NAME = "gen_ai.agent.name"

# GenAI Operation attributes
GEN_AI_OPERATION_NAME = "gen_ai.operation.name"
GEN_AI_SYSTEM = "gen_ai.system"

# GenAI Tool attributes
GEN_AI_TOOL_CALL_ID = "gen_ai.tool.call.id"
GEN_AI_TOOL_DESCRIPTION = "gen_ai.tool.description"
GEN_AI_TOOL_NAME = "gen_ai.tool.name"

# Error attributes
ERROR_TYPE = "error.type"


class GenAiOperationNameValues(Enum):
    """Enum for GenAI operation name values."""

    CHAT = "chat"
    CREATE_AGENT = "create_agent"
    EMBEDDINGS = "embeddings"
    EXECUTE_TOOL = "execute_tool"
    GENERATE_CONTENT = "generate_content"
    INVOKE_AGENT = "invoke_agent"
    TEXT_COMPLETION = "text_completion"


# Constant for system name
GENAI_SYSTEM_AUTOGEN = "autogen"


@contextmanager
def trace_tool_span(
    tool_name: str,
    *,
    tracer: Optional[trace.Tracer] = None,
    parent: Optional[Span] = None,
    tool_description: Optional[str] = None,
    tool_call_id: Optional[str] = None,
) -> Generator[Span, Any, None]:
    """Context manager to create a span for tool execution following the
    OpenTelemetry Semantic conventions for generative AI systems.

    See the GenAI semantic conventions documentation:
    `OpenTelemetry GenAI Semantic Conventions <https://opentelemetry.io/docs/specs/semconv/gen-ai/>`__

    .. warning::

        The GenAI Semantic Conventions are still in incubation and
        subject to changes in future releases.


    Args:
        tool_name (str): The name of the tool being executed.
        tracer (Optional[trace.Tracer]): The tracer to use for creating the span.
        parent (Optional[Span]): The parent span to link this span to.
        tool_description (Optional[str]): A description of the tool.
        tool_call_id (Optional[str]): A unique identifier for the tool call.
    """
    if tracer is None:
        tracer = trace.get_tracer("autogen-core")
    span_attributes = {
        GEN_AI_OPERATION_NAME: GenAiOperationNameValues.EXECUTE_TOOL.value,
        GEN_AI_SYSTEM: GENAI_SYSTEM_AUTOGEN,
        GEN_AI_TOOL_NAME: tool_name,
    }
    if tool_description is not None:
        span_attributes[GEN_AI_TOOL_DESCRIPTION] = tool_description
    if tool_call_id is not None:
        span_attributes[GEN_AI_TOOL_CALL_ID] = tool_call_id
    with tracer.start_as_current_span(
        f"{GenAiOperationNameValues.EXECUTE_TOOL.value} {tool_name}",
        kind=SpanKind.INTERNAL,
        context=trace.set_span_in_context(parent) if parent else None,
        attributes=span_attributes,
    ) as span:
        try:
            yield span
        except Exception as e:
            # Set the exception details on the span if an error occurs
            span.record_exception(e)
            span.set_status(trace.Status(trace.StatusCode.ERROR, str(e)))
            span.set_attribute(ERROR_TYPE, type(e).__name__)
            raise


@contextmanager
def trace_create_agent_span(
    agent_name: str,
    *,
    tracer: Optional[trace.Tracer] = None,
    parent: Optional[Span] = None,
    agent_id: Optional[str] = None,
    agent_description: Optional[str] = None,
) -> Generator[Span, Any, None]:
    """Context manager to create a span for agent creation following the
    OpenTelemetry Semantic conventions for generative AI systems.

    See the GenAI semantic conventions documentation:
    `OpenTelemetry GenAI Semantic Conventions <https://opentelemetry.io/docs/specs/semconv/gen-ai/>`__

    .. warning::

        The GenAI Semantic Conventions are still in incubation and
        subject to changes in future releases.

    Args:
        agent_name (str): The name of the agent being created.
        tracer (Optional[trace.Tracer]): The tracer to use for creating the span.
        parent (Optional[Span]): The parent span to link this span to.
        agent_id (Optional[str]): The unique identifier for the agent.
        agent_description (Optional[str]): A description of the agent.
    """
    if tracer is None:
        tracer = trace.get_tracer("autogen-core")
    span_attributes = {
        GEN_AI_OPERATION_NAME: GenAiOperationNameValues.CREATE_AGENT.value,
        GEN_AI_SYSTEM: GENAI_SYSTEM_AUTOGEN,
        GEN_AI_AGENT_NAME: agent_name,
    }
    if agent_id is None:
        # Try to see if we can get the agent ID from the current context
        try:
            agent_id = str(AgentInstantiationContext.current_agent_id())
        except RuntimeError:
            agent_id = None
    if agent_id is not None:
        span_attributes[GEN_AI_AGENT_ID] = agent_id
    if agent_description is not None:
        span_attributes[GEN_AI_AGENT_DESCRIPTION] = agent_description
    with tracer.start_as_current_span(
        f"{GenAiOperationNameValues.CREATE_AGENT.value} {agent_name}",
        kind=SpanKind.CLIENT,
        context=trace.set_span_in_context(parent) if parent else None,
        attributes=span_attributes,
    ) as span:
        try:
            yield span
        except Exception as e:
            # Set the exception details on the span if an error occurs
            span.record_exception(e)
            span.set_status(trace.Status(trace.StatusCode.ERROR, str(e)))
            span.set_attribute(ERROR_TYPE, type(e).__name__)
            raise


@contextmanager
def trace_invoke_agent_span(
    agent_name: str,
    *,
    tracer: Optional[trace.Tracer] = None,
    parent: Optional[Span] = None,
    agent_id: Optional[str] = None,
    agent_description: Optional[str] = None,
) -> Generator[Span, Any, None]:
    """Context manager to create a span for invoking an agent following the
    OpenTelemetry Semantic conventions for generative AI systems.

    See the GenAI semantic conventions documentation:
    `OpenTelemetry GenAI Semantic Conventions <https://opentelemetry.io/docs/specs/semconv/gen-ai/>`__

    .. warning::

        The GenAI Semantic Conventions are still in incubation and
        subject to changes in future releases.

    Args:
        agent_name (str): The name of the agent being invoked.
        tracer (Optional[trace.Tracer]): The tracer to use for creating the span.
        parent (Optional[Span]): The parent span to link this span to.
        agent_id (Optional[str]): The unique identifier for the agent.
        agent_description (Optional[str]): A description of the agent.
    """
    if tracer is None:
        tracer = trace.get_tracer("autogen-core")
    span_attributes = {
        GEN_AI_OPERATION_NAME: GenAiOperationNameValues.INVOKE_AGENT.value,
        GEN_AI_SYSTEM: GENAI_SYSTEM_AUTOGEN,
        GEN_AI_AGENT_NAME: agent_name,
    }
    if agent_id is not None:
        span_attributes[GEN_AI_AGENT_ID] = agent_id
    if agent_description is not None:
        span_attributes[GEN_AI_AGENT_DESCRIPTION] = agent_description
    with tracer.start_as_current_span(
        f"{GenAiOperationNameValues.INVOKE_AGENT.value} {agent_name}",
        kind=SpanKind.CLIENT,
        context=trace.set_span_in_context(parent) if parent else None,
        attributes=span_attributes,
    ) as span:
        try:
            yield span
        except Exception as e:
            # Set the exception details on the span if an error occurs
            span.record_exception(e)
            span.set_status(trace.Status(trace.StatusCode.ERROR, str(e)))
            span.set_attribute(ERROR_TYPE, type(e).__name__)
            raise
